/*
 * name: mdialog.cpp
 * description: main dialog window
 * author: tilman jochems
 * (c)11/2004, midi-formatique, rivesaltes
 */
#include <vcl.h>
#include <math.h>
#pragma hdrstop
#include "Tools.h"
#include "Transmitter.h"
#include "DataBase.h"
#include "MDialog.h"
#pragma package(smart_init)
#pragma resource "*.dfm"
TFMDialog *FMDialog;

#define FREQ                100        //frequency of thread
#define NRMESSAGES           25        //size of message queue

//construction
__fastcall TFMDialog::TFMDialog(TComponent* Owner):TForm(Owner){
 TTable *rem;
 TSpeedButton *sb;

 //initialize variables
 ActRem=0;
 OffsetX=OffsetY=0;
 MouseIn=false;
 KeyDown=false;

 //initialize buttons for remotes
 rem=Data->TRemotes;
 for(NrRem=0,rem->First();
     NrRem<5 && !rem->Eof;
     NrRem++,rem->Next()){
  sb=(TSpeedButton *)FindComponent("SB" + AnsiString(NrRem));
  SBRem[NrRem]=sb;
  sb->Caption=Data->TRemotesCaption->Value;
  sb->Hint=Data->TRemotesName->Value;
  sb->Tag=NrRem;
  sb->Visible=true;
 }//for

 //create critical section and threads (suspended mode)
 CSection=new TCriticalSection();      //critical section (global.cpp)
 Transmitter=new TTransmitter(true);   //thread (suspended)

 //create and initialize message queue
 MQueue=new TMessQueue(NRMESSAGES,MDIALOG_NRHANDLERS);
}//TFMDialog

//destruction
__fastcall TFMDialog::~TFMDialog(){
 delete Transmitter;
 delete MQueue;
 delete CSection;
}//~TFMDialog

//when coming up
void __fastcall TFMDialog::FormShow(TObject *Sender)
{
 //initialize hardware
#ifdef PIO
 Transmitter->MQueue->Post(TRANSMITTER_STARTDRV,0,0);
#endif
#ifdef COM
 //get name of port
 TLocateOptions opt;
 static AnsiString port;
 static int addr;
 if(!Data->TConstants->Locate("Name","Port",opt))
  FatalError("can't get port name");
 port=Data->TConstantsValue->Value;
 if(!Data->TConstants->Locate("Name","PAddr",opt))
  FatalError("can't get address of port");
 addr=Data->TConstantsValue->AsInteger;
 Transmitter->MQueue->Post(TRANSMITTER_STARTDRV,(int)(&port),addr);
#endif

 //show first control
 SBClick(SBRem[0]);

 //everybody is ready -> here we go
 Transmitter->Resume();
 Application->OnIdle=Execute;
}//FormShow

//when closing the application
void __fastcall TFMDialog::FormHide(TObject *Sender)
{
 //stop threads
 Transmitter->Terminate();
 Transmitter->WaitFor();
}//FormHide

//main loop of thread (called by application when idle)
void __fastcall TFMDialog::Execute(TObject *Sender, bool &Done){
 uint time,delta;                      //timing of thread
 static uint lasttime=0;
 MQ_MSG msg;                           //incomming messages

 //dispatch all incoming intern messages until timeout
 do{
  time=GetTickCount();
  delta=INTM(time,lasttime);
  delta=(delta<FREQ) ?  FREQ-delta : 0;
  if(MQueue->Get(&msg,delta)){MQueue->Dispatch(msg);continue;}
 }while(delta!=0);
 lasttime=GetTickCount();

 //call idlehandler and execute again
 IdleHandler();
 Done=false;
}//Execute

//idle handler (called with FREQ)
void TFMDialog::IdleHandler(){
}//IdleHandler

//draw button around control (for testing the database)
void TFMDialog::DrawButton(int form,int x,int y,int w,int h,TColor col,bool fill){
 TCanvas *canvas=IRemote->Canvas;

 canvas->Brush->Color=col;
 canvas->Brush->Style= fill ? bsSolid : bsClear;
 canvas->Pen->Color=clLime;
 switch(form){
  case ELLIPSE:
   canvas->Ellipse(x-w,y-h,x+w,y+h);
  break;
  case RECTANGLE:
   canvas->Rectangle(x-w,y-h,x+w,y+h);
  break;
 }//switch
}//DrawButton

//change remote control (speed button in top of form)
void __fastcall TFMDialog::SBClick(TObject *Sender)
{
 TLocateOptions opt;
 TQuery *query;
 AnsiString iname;
 TSpeedButton *button;

 //get speed button which was clicked
 button=(TSpeedButton *)Sender;
 button->Down=true;
 ActRem=button->Tag;

 //locate remote in data base
 if(!Data->TRemotes->Locate("Remote",ActRem,opt)){
  NotFatalError("can't locate remote control in database");
  return;
 };//if

 //load image for remote control
 iname=DATADIR;
 iname+="/" + Data->TRemotesImage->Value;
 IRemote->Picture->LoadFromFile(iname);
 OffsetX=(IRemote->Width-IRemote->Picture->Width)/2;
 OffsetY=(IRemote->Height-IRemote->Picture->Height)/2;

 //set table with codes for selected remote control
 query=Data->QCode;
 query->Close();
 query->Params->Items[0]->Value=ActRem;
 query->Prepare();
 query->Open();

 //inform transmitter about used protocol
 Protocol=Data->TRemotesProtocol->Value;
 Transmitter->MQueue->Post(TRANSMITTER_SETPROTO,Protocol,0);

 //show sensitive zones
#if DEBUG==1
 for(query->First();!query->Eof;query->Next()){
 DrawButton(Data->QCodeForme->Value,
            Data->QCodeX->Value,Data->QCodeY->Value,
            Data->QCodeW->Value,Data->QCodeH->Value,
            clWhite,false);
 }//for
#endif
}//SBClick

//mouse moved on picture of remote control
void __fastcall TFMDialog::IRemoteMouseMove(TObject *Sender,
      TShiftState Shift, int X, int Y)
{
 //correction of coordinates
 X-=OffsetX;
 Y-=OffsetY;

#if DEBUG==0
 int x,y,w,h,dx,dy;
 TQuery *query;

 //test if mouse on key
 query=Data->QCode;
 for(query->First();!query->Eof;query->Next()){
  x=Data->QCodeX->Value;
  y=Data->QCodeY->Value;
  w=Data->QCodeW->Value;
  h=Data->QCodeH->Value;
  dx=abs(x-X);
  dy=abs(y-Y);
  switch(Data->QCodeForme->Value){
   case ELLIPSE:
    MouseIn=pow((double)dx/w,2)+pow((double)dy/h,2)<=2;
   break;
   case RECTANGLE:
    MouseIn=(dx<w && dy<h);
   break;
  }//switch

  //stop loop if button found
  if(MouseIn) break;
 }//for

 //if mouse in button, show command and save values
 if(MouseIn){
  Address=Data->QCodeAddress->Value;
  Code=Data->QCodeCode->Value;
  LMessage->Caption=Data->QCodeDescription->Value;
 }else{
  LMessage->Caption="";
 }//else
#else
 //show coordinates of mouse pointer
 char buf[32];
 wsprintf(buf,"x: %.3d  y: %.3d",X,Y);
 LMessage->Caption=buf;
#endif
}//IRemoteMouseMove

//mouse on button: send command to transmitter
void __fastcall TFMDialog::IRemoteMouseDown(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
 static int send[2];
 if(!MouseIn) return;
 LMessage->Color=clLime;
 send[0]=Address;
 send[1]=Code;
 Transmitter->MQueue->Post(TRANSMITTER_COMSEND,true,(int)send);
}//IRemoteMouseDown

//mouse up: end sending
void __fastcall TFMDialog::IRemoteMouseUp(TObject *Sender,
      TMouseButton Button, TShiftState Shift, int X, int Y)
{
 LMessage->Color=Color;
 Transmitter->MQueue->Post(TRANSMITTER_COMSEND,false,0);
}//IRemoteMouseUp

//execute shortcuts from keyboard
void __fastcall TFMDialog::FormKeyDown(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
#if DEBUG==0
 TQuery *query;
 static int send[2];

 //keys for application
 if(Key==115 &&                        //ctrl-F4
    Shift.Contains(ssCtrl)){
  Application->Minimize();
  return;
 }//if

 //test if key allready pressed
 if(KeyDown) return;
 KeyDown=true;

 //search for key in table
 query=Data->QCode;
 for(query->First();!query->Eof;query->Next())
  if(Data->QCodeShortCut->Value==Key) goto send;

 //no shortcut defined
 LMessage->Caption="";
 return;

 //send command if shortcut is defined
send:
 send[0]=Data->QCodeAddress->Value;
 send[1]=Data->QCodeCode->Value;
 LMessage->Caption=Data->QCodeDescription->Value;
 LMessage->Color=clLime;
 Transmitter->MQueue->Post(TRANSMITTER_COMSEND,true,(int)send);
#else
 //show coordinates of mouse pointer
 char buf[32];
 wsprintf(buf,"key: %d %d %d",Key,Shift.Contains(ssShift),Shift.Contains(ssCtrl));
 LMessage->Caption=buf;
#endif
}//FormKeyDown

//key up: end sending
void __fastcall TFMDialog::FormKeyUp(TObject *Sender, WORD &Key,
      TShiftState Shift)
{
 KeyDown=false;
 LMessage->Color=Color;
 Transmitter->MQueue->Post(TRANSMITTER_COMSEND,false,0);
}//FormKeyUp

//test by sending continously
void __fastcall TFMDialog::TTestTimer(TObject *Sender)
{
#if DEBUG==1
 static int send[2],i1=0;
 send[0]=24225;
 send[1]=8;
 //LMessage->Caption=++i1;
 //Transmitter->MQueue->Post(TRANSMITTER_COMSEND,true,(int)send);
#endif
}
//---------------------------------------------------------------------------

