/*
 * name: messqueue.cpp
 * description: message queue (send, post, get, dispatch,...)
 * -enables easy and asynchronous communication between threads
 * -a thread can post/send a message to another
 * -this one will read the message and dispatch it (all a handler function)
 * author: tilman jochems
 * (c)10/2004, midi-formatique, rivesaltes
 */
#include <vcl.h>
#pragma hdrstop
#include "Tools.h"
#include "MessQueue.h"

#define RWNR     20                     //20x100ms=timeout for qread/qwrite
#define RWSLEEP 100

//construction
TMessQueue::TMessQueue(int qs,int hs){

 //save argguments
 QSize=qs;                             //max number of messages in queue
 HSize=hs;                             //number of handlers

 //create new queue of messages
 Queue=new MQ_MSG[QSize];
 if(Queue==NULL) FatalError("TMessQueue : memory allocation");
 QReadPos=QWritePos=0;

 //initialize table of callbacks
 Handlers=new MQ_CALLBACK[HSize];
 if(Handlers==NULL) FatalError("TMessQueue : memory allocation");
 memset(Handlers,0,HSize*sizeof(MQ_CALLBACK));
}//TMessQueue

//destruction
TMessQueue::~TMessQueue(){
 delete Queue;
 delete Handlers;
}//~TMessQueue

//test if message queue is empty
bool TMessQueue::QEmpty(){
 bool res;
 ECS
 res=(QReadPos==QWritePos);
 LCS
 return res;
}//QEmpty

//write message to the queue
void TMessQueue::QWrite(MQ_MSG msg){
 int i1,np;

 //wait until one place empty for writing
 for(i1=0;i1<RWNR;i1++){
  ECS
  np=(QWritePos+1)%QSize;
  if(np==QReadPos){LCS Sleep(RWSLEEP); continue;}
  Queue[QWritePos]=msg;
  QWritePos=np;
  LCS
  return;
 }//for

 //try to continue without writing msg
 NotFatalError("QWrite : the queue is full");
}//QWrite

//read message from queue
MQ_MSG TMessQueue::QRead(){
 MQ_MSG msg;
 int i1;

 //wait until receiving a message or timeout
 for(i1=0;i1<RWNR;i1++){
  ECS
  if(QEmpty()){LCS Sleep(RWSLEEP);continue;}
  msg=Queue[QReadPos];
  QReadPos=(QReadPos+1)%QSize;
  LCS
  return msg;
 }//for

 //try to continue without a message (does this happen ?)
 NotFatalError("QRead : the queue is empty");
 msg.nr=-1;
 return msg;
}//QRead

//add a callback to the dispatch table
void TMessQueue::AddHandler(int fct,MQ_CALLBACK callback){
 if(fct>=HSize) FatalError("AddHandler: the table is too small");
 Handlers[fct]=callback;
}//AddHandler

//dispatch message (call handler)
void TMessQueue::Dispatch(MQ_MSG msg){
 int res;
 if(msg.nr>=HSize ||
    Handlers[msg.nr]==NULL) FatalError("Dispatch: not valid message");

 //call function
 res=(*Handlers[msg.nr])(msg.arg1,msg.arg2);

 //if message was send (not posted), save result and set event
 if(msg.send){
  *msg.res=res;
  SetEvent(msg.event);
 }//if
}//Dispatch

//queue a message and wait until end of execution
int  TMessQueue::Send(int fct,int arg1,int arg2){
 MQ_MSG msg;
 int res;

 //fill in message structure and queue it
 msg.nr   =fct;
 msg.arg1 =arg1;
 msg.arg2 =arg2;
 msg.send =true;
 msg.res  =&res;
 msg.event=CreateEvent(NULL,false,false,NULL);

 //queue message and wait for execution
 QWrite(msg);
 WaitForSingleObject(msg.event,INFINITE);

 //free and return
 CloseHandle(msg.event);
 return res;
}//Send

//queue a message and return immediately
void TMessQueue::Post(int fct,int arg1,int arg2){
 MQ_MSG msg;

 //fill in message structure and queue it
 msg.nr   =fct;
 msg.arg1 =arg1;
 msg.arg2 =arg2;
 msg.send =false;
 msg.res  =NULL;
 msg.event=NULL;

 //queue message and return immediately
 QWrite(msg);
}//Post

//get message (blocking with timeout)
//if message -> read it and return true,
//else       -> block until timeout and return false
bool TMessQueue::Get(MQ_MSG *msg,uint timeout){
 uint now,delta;
 uint first=GetTickCount();

 //loop until timeout
 while(true){
  //test if new message arrived
  if(!QEmpty()){
   *msg=QRead();
   return true;                        //new message read
  }//if

  //let other threads work until timeout
  now=GetTickCount();
  delta=INTM(now,first);
  if(delta>=timeout) break;
  Sleep(1);
 }//while

 return false;                         //timeout error
}//Get

