/*
 * name: transmitter.cpp
 * description: transmitting remote control commands
 * author: tilman jochems
 * (c)11/2004, midi-formatique, rivesaltes
 */
#include <vcl.h>
#include <stdio.h>
#pragma hdrstop
#include "Tools.h"
#include "MDialog.h"
#include "Transmitter.h"
#pragma package(smart_init)

//frequency of the thread
#define FREQ                 50        //frequency of thread in ms
#define NRMESSAGES           25        //size of message queue

//carte PIO, (port a bit 0 = input, port b bit 0 = output)
#ifdef PIO
bool PioInit();                        //initialize card
void PioReset();                       //reset card
#define BASE    0x0210
#define PIO_A  (BASE + 0)		       //port registers
#define PIO_B  (BASE + 1)
#define PIO_C  (BASE + 2)
#define PIO_CR (BASE + 3)			   //control register
#define DrvInit()  PioInit()
#define DrvReset() PioReset()
#define DrvOut(a)  outp(PIO_B,(a) & 1)
#define DrvInp()   (inp(PIO_A) & 1)
#endif

//rs323 (only output)
#ifdef COM
//#define ZE5                            //zero is 5V
AnsiString Port;                       //name of com port
ushort PAddr=0;                        //address of com port
HANDLE HCom=NULL;                      //handle to com
bool ComInit();                        //initialize com
void ComReset();                       //reset com
void ComOut(int val);
#define DrvInit()  ComInit()
#define DrvReset() ComReset()
#define DrvOut(a)  ComOut(a);
#define DrvInp()   1
#endif

#define TIMEERROR -1                   //time error when reading (result)

//timing of protocolas
#define RC5_TIME   889                 //time base T
#define RC5_FREQ   36.0                //frequence in khz

#define GRU_TIME   500                 //time base T
#define GRU_FREQ   36.0                //frequence in khz

#define YAM_TIME   550                 //time base T
#define YAM_FREQ   36.0                //frequence in khz (39.2?)

//global variables
TTransmitter *Transmitter;

//construction
__fastcall TTransmitter::TTransmitter(bool CreateSuspended)
    : TThread(CreateSuspended)
{
 //initialize variables
 Protocol=PROTO_RC5;
 Sending=false;
 SendTime=0;

 //initialize performance counter
 if(!PCntInit()) FatalError("can't locate performance counter");

 //create and initialize message queue
 MQueue=new TMessQueue(NRMESSAGES,TRANSMITTER_NRHANDLERS);
 MQueue->AddHandler(TRANSMITTER_COMSEND,ComSend);
 MQueue->AddHandler(TRANSMITTER_SETPROTO,SetProto);
 MQueue->AddHandler(TRANSMITTER_STARTDRV,StartDrv);

 //receiving/transmitting is time critical
 Priority=tpTimeCritical;
}//TTransmitter

//destruction
__fastcall TTransmitter::~TTransmitter(){
 DrvReset();
 delete MQueue;
}//~TTransmitter

//main loop of thread
void __fastcall TTransmitter::Execute()
{
 MQ_MSG msg;
 uint time,delta,lasttime=0;

 //loop until terminated
 while(!Terminated){

  //dispatch all incomming messages until FREQ ms have passed
  do{
   time=GetTickCount();
   delta=INTM(time,lasttime);
   delta=(delta<FREQ) ?  FREQ-delta : 0;
   if(MQueue->Get(&msg,delta)){MQueue->Dispatch(msg);continue;}
  }while(delta!=0);
  lasttime=GetTickCount();

  //send code
  if(Sending){
   switch(Protocol){
    case PROTO_RC5:     Rc5Send(); break;
    case PROTO_GRUNDIG: GruSend(); break;
    case PROTO_YAMAHA:  YamSend(); break;
   }//switch
  }//if
 }//while
}//Execute

/*
 * message handlers
 */
//start/stop sending mode
int TTransmitter::ComSend(int on,int code){
 int *send=(int *)code;

 Transmitter->Sending=on;                 //switch permanent sending on/off
 if(on){                                  //if on, save code
  Transmitter->Toggle^=1;                 //for rc5
  Transmitter->Address=send[0];
  Transmitter->Code=send[1];
 }//if

 //if Grunding, send start/stop frame
 if(Transmitter->Protocol==PROTO_GRUNDIG) Transmitter->GruSend(true);

 //if Yamaha, send code frame once at start
 if(Transmitter->Protocol==PROTO_YAMAHA && on) Transmitter->YamSend(true);

 return true;
}//ComSend

//set protocol of communication
int TTransmitter::SetProto(int proto,int){
 Transmitter->Protocol=proto;
 return true;
}//SetProto

//start driver
int TTransmitter::StartDrv(int par1,int par2){
#ifdef COM
 Port=*((AnsiString *)par1);
 PAddr=(ushort)par2;
#endif
 //restart driver
 DrvReset();
 return DrvInit();
}//int


/*
 * pio card, based on a 8255
 */
#ifdef PIO
//initialize card
bool PioInit(){
 outp(PIO_CR,0x99);		   //A as input, B as output
 outp(PIO_B,1);            //B high output
 return (inp(PIO_B)==1);
}//PioInit

//all ports input
void PioReset(){
 outp(PIO_CR,0x9B);        //all to inputs
}//PioReset
#endif

/*
 * rs232
 */
#ifdef COM

//initialize card
bool ComInit(){
 try{
  DCB dcb;

  //open communicatin device
  //GENERIC_READ | GENERIC_WRITE
  HCom=CreateFile(Port.c_str(),0,0,NULL,OPEN_EXISTING,0,NULL);
  if(HCom==INVALID_HANDLE_VALUE) throw "can't open com port";

  //configure com port (set RTS)
  if(!GetCommState(HCom,&dcb)) throw "can't get com state";
#ifdef ZE5
  dcb.fRtsControl=RTS_CONTROL_ENABLE;
#else
  dcb.fRtsControl=RTS_CONTROL_DISABLE;
#endif
  if(!SetCommState(HCom,&dcb)) throw "can't set com state";
 }//try
 catch(char *mess){
  FatalError(mess);
 }//catch
 return true;
}//ComInit

//all ports input
void ComReset(){
 if(HCom!=NULL) CloseHandle(HCom);
 HCom=NULL;
}//ComReset

//send 1 bit to RTS line
void ComOut(int val){
 ushort addr=(ushort)(PAddr+4);        //modem control register
 uchar mcr;
 mcr=inp(addr);
#ifdef ZE5
 if(val) mcr|=2;                       //bit 2: rts line
 else    mcr&=~2;
#else
 if(val) mcr&=~2;                      //bit 2: rts line
 else    mcr|=2;
#endif
 outp(addr,mcr);
}//ComOut
#endif

/*
 * functions for all protocols
 */

//send all bits in a buffer, dt is time for 1 bit
void TTransmitter::SendBits(int *bits,int nr,float freq,uint dt){
 uint64 time=PCntTicks();              //start of sending
 int nrp=dt*freq/1000;                 //number of pulses for 1 bit
 int i1,i2;
 uint t1,t2;

 //send all bits
 for(i1=0;i1<nr;i1++){

  //bit==1 is space, don't send
  if(bits[i1]==1) continue;

  //wait for start of bit
  t1=i1*dt;
  while(PCntDiff(time)<t1);

  //send pulses with freq (5us off, rest on)
  for(i2=0;i2<nrp;i2++){
   t2=t1+i2*1000/freq;                 //start of pulse
   while(PCntDiff(time)<t2);
   DrvOut(0);
   while(PCntDiff(time)<t2+7);         //end of pulse
   DrvOut(1);
  }//for
 }//for
}//SendBits

/*
 * RC-5 protocol
 * f=36khz T=889us
 * frame= start + 13 bits signal + 50 bits pause
 * start=T 0 + T 1 (=0)
 * bits: !c6 tO a4 a3 a2 a1 a0 c5 c4 c3 c2 c1 c0
 * to=toggle bit (when key pressed again)
 * bit: 1 0 is 1, 0 1 is 0
 */
//send code (14 bits), 1=1->0, 0=0->1
void TTransmitter::Rc5Send(){
 int i1,val,bits[28];

 //transform address/code into bits
 bits[0]=1;bits[1]=0;                 //start bit (=1)
 val=(Code >> 7) & 1;                 //c6
 bits[2]=!val;bits[3]=val;            //!c6 (=1 for older RC5)
 bits[4]=Toggle;bits[5]=!Toggle;      //toggle bit
 for(i1=0;i1<=4;i1++){                //address (a0-a4)
  val=(Address >> i1) & 1;
  bits[14-i1*2]=val;
  bits[15-i1*2]=!val;
 }//for
 for(i1=0;i1<=5;i1++){                //command (c0-c5)
  val=(Code >> i1) & 1;
  bits[26-i1*2]=val;
  bits[27-i1*2]=!val;
 }//for

 //wait for start of frame and send
 if(PCntDiff(SendTime)<30000) return;  //pause between signals : 89ms
 while(PCntDiff(SendTime)<113778);     //length of 1 cycle = 64 bits
 SendTime=PCntTicks();                 //start of cycle
 SendBits(bits,28,RC5_FREQ,RC5_TIME);         //send now
}//Rc5Send

/*
 * grundig protocol
 * f=36khz T=500us
 * start + 10 bits signal + 110 bits pause
 * start=T 0 + T 1 (=0)
 * bits: c0 c1 c2 c3 c4 c5 c6 a0 a1 a2
 * bit: 1 0 is 0, 0 1 is 1
 * key pressed or key relaxed, signal=1111111 111 (7F 07)
 */
//send code
void TTransmitter::GruSend(bool startstop){
 int i1,val,bits[26];

 //send startstop or code
 bits[0]=0;                             //start sequence
 for(i1=1;i1<=5;i1++) bits[i1]=1;
 if(startstop){
  for(i1=0;i1<=9;i1++){                 //command and address
   bits[6+2*i1]  =0;
   bits[6+2*i1+1]=1;
  }//for
 }else{
  for(i1=0;i1<=6;i1++){                 //command (c0-c6)
   val=(Code >> i1) & 1;
   bits[6+2*i1]  =!val;
   bits[6+2*i1+1]= val;
  }//for
  for(i1=0;i1<=2;i1++){                 //address (a0-a2)
   val=(Address >> i1) & 1;
   bits[20+2*i1]  =!val;
   bits[20+2*i1+1]= val;
  }//for
 }//else

 //wait for start of frame and send
 if(!startstop &&
    PCntDiff(SendTime)<30000) return;  //pause between signals : 102ms
 while(PCntDiff(SendTime)<115000);     //length of 1 cycle : 115ms
 SendTime=PCntTicks();                 //start of cycle
 SendBits(bits,26,GRU_FREQ,GRU_TIME);//send now
}//GruSend

/*
 * yamaha protocol
 * f=39.2khz T=550us
 * start + 32 bits signal + end
 * start=16*T 0 + 8*T 1
 * end=1*T 0 + 1*T 1
 * signal only send once, then repeat stop frame
 * stop=16*T 0 + 4*T 1 + 1*T 0 + 1*T 1
 * 105ms pause between start of headers
 * bits: 16bits address, 8bit command, 0xff ^ command
 * bit: 0 1 1 1 is 1, 0 1 is 0
 */
//send code
void TTransmitter::YamSend(bool code){
 int i1,count,val,bits[154];

 //write value 0 or 1
#define w0(a,b) (b[a++]=0,b[a++]=1)
#define w1(a,b) (b[a++]=0,b[a++]=1,b[a++]=1,b[a++]=1)

 //send code or stop frame
 if(code){
  //header
  for(count=i1=0;i1<16;i1++) bits[count++]=0;
  for(i1=0;i1<8;i1++)        bits[count++]=1;

  //address
  for(i1=15;i1>=0;i1--)
   if((Address>>i1) & 1) w1(count,bits); else w0(count,bits);

  //code and ~code
  for(val=Code,i1=7;i1>=0;i1--)
   if((val>>i1) & 1) w1(count,bits); else w0(count,bits);
  for(val=~Code,i1=7;i1>=0;i1--)
   if((val>>i1) & 1) w1(count,bits); else w0(count,bits);
  w0(count,bits);

 }else{
  for(count=i1=0;i1<16;i1++) bits[count++]=0;
  for(i1=0;i1<4;i1++)        bits[count++]=1;
  w0(count,bits);
 }//for

 //wait for start of frame and send
 if(!code &&
    PCntDiff(SendTime)<30000) return;
 while(PCntDiff(SendTime)<105000);     //length of 1 cycle : 105ms
 SendTime=PCntTicks();                 //start of cycle
 SendBits(bits,count,YAM_FREQ,YAM_TIME);//send now
 
#undef w0
#undef w1
}//YamSend


